@@ -313,6 +313,7 @@ class ArrayStrategy(st.SearchStrategy):
     def __init__(
         self, *, xp, api_version, elements_strategy, dtype, shape, fill, unique
     ):
+        super().__init__()
         self.xp = xp
         self.elements_strategy = elements_strategy
         self.dtype = dtype
@@ -499,9 +500,6 @@ def _arrays(
       >>> xps.arrays(xp, xp.int8, 3, elements={"min_value": 10}).example()
       Array([125, 13, 79], dtype=int8)
 
-    Refer to :doc:`What you can generate and how <data>` for passing
-    your own elements strategy.
-
     .. code-block:: pycon
 
       >>> xps.arrays(xp, xp.float32, 3, elements=floats(0, 1, width=32)).example()
@@ -1037,24 +1035,24 @@ def make_strategies_namespace(
                 f_args += f", api_version='{self.api_version}'"
             return f"make_strategies_namespace({f_args})"
 
-    kwargs = dict(
-        name=xp.__name__,
-        api_version=api_version,
-        from_dtype=from_dtype,
-        arrays=arrays,
-        array_shapes=array_shapes,
-        scalar_dtypes=scalar_dtypes,
-        boolean_dtypes=boolean_dtypes,
-        real_dtypes=real_dtypes,
-        numeric_dtypes=numeric_dtypes,
-        integer_dtypes=integer_dtypes,
-        unsigned_integer_dtypes=unsigned_integer_dtypes,
-        floating_dtypes=floating_dtypes,
-        valid_tuple_axes=valid_tuple_axes,
-        broadcastable_shapes=broadcastable_shapes,
-        mutually_broadcastable_shapes=mutually_broadcastable_shapes,
-        indices=indices,
-    )
+    kwargs = {
+        "name": xp.__name__,
+        "api_version": api_version,
+        "from_dtype": from_dtype,
+        "arrays": arrays,
+        "array_shapes": array_shapes,
+        "scalar_dtypes": scalar_dtypes,
+        "boolean_dtypes": boolean_dtypes,
+        "real_dtypes": real_dtypes,
+        "numeric_dtypes": numeric_dtypes,
+        "integer_dtypes": integer_dtypes,
+        "unsigned_integer_dtypes": unsigned_integer_dtypes,
+        "floating_dtypes": floating_dtypes,
+        "valid_tuple_axes": valid_tuple_axes,
+        "broadcastable_shapes": broadcastable_shapes,
+        "mutually_broadcastable_shapes": mutually_broadcastable_shapes,
+        "indices": indices,
+    }
 
     if api_version > "2021.12":
 
