@@ -8,50 +8,51 @@
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
 # obtain one at https://mozilla.org/MPL/2.0/.
 
+import abc
 import inspect
 import math
 from random import Random
-from typing import Any, Dict
+from typing import Any
 
 import attr
 
 from hypothesis.control import should_note
+from hypothesis.internal.conjecture.data import ConjectureData
 from hypothesis.internal.reflection import define_function_signature
 from hypothesis.reporting import report
-from hypothesis.strategies._internal.core import (
-    binary,
-    lists,
-    permutations,
-    sampled_from,
-)
+from hypothesis.strategies._internal.core import lists, permutations, sampled_from
 from hypothesis.strategies._internal.numbers import floats, integers
 from hypothesis.strategies._internal.strategies import SearchStrategy
 
 
-class HypothesisRandom(Random):
+class HypothesisRandom(Random, abc.ABC):
     """A subclass of Random designed to expose the seed it was initially
     provided with."""
 
-    def __init__(self, note_method_calls):
-        self.__note_method_calls = note_method_calls
+    def __init__(self, *, note_method_calls: bool) -> None:
+        self._note_method_calls = note_method_calls
 
     def __deepcopy__(self, table):
         return self.__copy__()
 
-    def __repr__(self):
-        raise NotImplementedError
-
+    @abc.abstractmethod
     def seed(self, seed):
         raise NotImplementedError
 
+    @abc.abstractmethod
     def getstate(self):
         raise NotImplementedError
 
+    @abc.abstractmethod
     def setstate(self, state):
         raise NotImplementedError
 
+    @abc.abstractmethod
+    def _hypothesis_do_random(self, method, kwargs):
+        raise NotImplementedError
+
     def _hypothesis_log_random(self, method, kwargs, result):
-        if not (self.__note_method_calls and should_note()):
+        if not (self._note_method_calls and should_note()):
             return
 
         args, kwargs = convert_kwargs(method, kwargs)
@@ -60,9 +61,6 @@ class HypothesisRandom(Random):
         )
         report(f"{self!r}.{method}({argstr}) -> {result!r}")
 
-    def _hypothesis_do_random(self, method, kwargs):
-        raise NotImplementedError
-
 
 RANDOM_METHODS = [
     name
@@ -110,7 +108,7 @@ def _randbelow(self, n: int) -> int:  # type: ignore
 STUBS = {f.__name__: f for f in [getrandbits, random, _randbelow]}
 
 
-SIGNATURES: Dict[str, inspect.Signature] = {}
+SIGNATURES: dict[str, inspect.Signature] = {}
 
 
 def sig_of(name):
@@ -171,9 +169,6 @@ def state_for_seed(data, seed):
     return state
 
 
-UNIFORM = floats(0, 1)
-
-
 def normalize_zero(f: float) -> float:
     if f == 0.0:
         return 0.0
@@ -184,17 +179,17 @@ def normalize_zero(f: float) -> float:
 class ArtificialRandom(HypothesisRandom):
     VERSION = 10**6
 
-    def __init__(self, note_method_calls, data):
+    def __init__(self, *, note_method_calls: bool, data: ConjectureData) -> None:
         super().__init__(note_method_calls=note_method_calls)
         self.__data = data
         self.__state = RandomState()
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "HypothesisRandom(generated data)"
 
-    def __copy__(self):
+    def __copy__(self) -> "ArtificialRandom":
         result = ArtificialRandom(
-            note_method_calls=self._HypothesisRandom__note_method_calls,
+            note_method_calls=self._note_method_calls,
             data=self.__data,
         )
         result.setstate(self.getstate())
@@ -233,8 +228,12 @@ class ArtificialRandom(HypothesisRandom):
 
         if method == "_randbelow":
             result = self.__data.draw_integer(0, kwargs["n"] - 1)
-        elif method in ("betavariate", "random"):
-            result = self.__data.draw(UNIFORM)
+        elif method == "random":
+            # See https://github.com/HypothesisWorks/hypothesis/issues/4297
+            # for numerics/bounds of "random" and "betavariate"
+            result = self.__data.draw(floats(0, 1, exclude_max=True))
+        elif method == "betavariate":
+            result = self.__data.draw(floats(0, 1))
         elif method == "uniform":
             a = normalize_zero(kwargs["a"])
             b = normalize_zero(kwargs["b"])
@@ -324,8 +323,8 @@ class ArtificialRandom(HypothesisRandom):
         elif method == "shuffle":
             result = self.__data.draw(permutations(range(len(kwargs["x"]))))
         elif method == "randbytes":
-            n = kwargs["n"]
-            result = self.__data.draw(binary(min_size=n, max_size=n))
+            n = int(kwargs["n"])
+            result = self.__data.draw_bytes(min_size=n, max_size=n)
         else:
             raise NotImplementedError(method)
 
@@ -398,7 +397,7 @@ def convert_kwargs(name, kwargs):
 
 class TrueRandom(HypothesisRandom):
     def __init__(self, seed, note_method_calls):
-        super().__init__(note_method_calls)
+        super().__init__(note_method_calls=note_method_calls)
         self.__seed = seed
         self.__random = Random(seed)
 
@@ -411,15 +410,15 @@ class TrueRandom(HypothesisRandom):
         args, kwargs = convert_kwargs(method, kwargs)
         return fn(*args, **kwargs)
 
-    def __copy__(self):
+    def __copy__(self) -> "TrueRandom":
         result = TrueRandom(
             seed=self.__seed,
-            note_method_calls=self._HypothesisRandom__note_method_calls,
+            note_method_calls=self._note_method_calls,
         )
         result.setstate(self.getstate())
         return result
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"Random({self.__seed!r})"
 
     def seed(self, seed):
@@ -433,12 +432,13 @@ class TrueRandom(HypothesisRandom):
         self.__random.setstate(state)
 
 
-class RandomStrategy(SearchStrategy):
-    def __init__(self, note_method_calls, use_true_random):
+class RandomStrategy(SearchStrategy[HypothesisRandom]):
+    def __init__(self, *, note_method_calls: bool, use_true_random: bool) -> None:
+        super().__init__()
         self.__note_method_calls = note_method_calls
         self.__use_true_random = use_true_random
 
-    def do_draw(self, data):
+    def do_draw(self, data: ConjectureData) -> HypothesisRandom:
         if self.__use_true_random:
             seed = data.draw_integer(0, 2**64 - 1)
             return TrueRandom(seed=seed, note_method_calls=self.__note_method_calls)
