@@ -11,12 +11,13 @@
 import copy
 import re
 import warnings
-from functools import lru_cache, partial
+from functools import cache, lru_cache, partial
 from typing import Optional
 
 from hypothesis.errors import HypothesisWarning, InvalidArgument
 from hypothesis.internal import charmap
-from hypothesis.internal.conjecture.data import COLLECTION_DEFAULT_MAX_SIZE
+from hypothesis.internal.conjecture.data import ConjectureData
+from hypothesis.internal.conjecture.providers import COLLECTION_DEFAULT_MAX_SIZE
 from hypothesis.internal.filtering import max_len, min_len
 from hypothesis.internal.intervalsets import IntervalSet
 from hypothesis.internal.reflection import get_pretty_function_description
@@ -31,7 +32,7 @@ from hypothesis.vendor.pretty import pretty
 
 
 # Cache size is limited by sys.maxunicode, but passing None makes it slightly faster.
-@lru_cache(maxsize=None)
+@cache
 def _check_is_single_character(c):
     # In order to mitigate the performance cost of this check, we use a shared cache,
     # even at the cost of showing the culprit strategy in the error message.
@@ -43,10 +44,13 @@ def _check_is_single_character(c):
     return c
 
 
-class OneCharStringStrategy(SearchStrategy):
+class OneCharStringStrategy(SearchStrategy[str]):
     """A strategy which generates single character strings of text type."""
 
-    def __init__(self, intervals, force_repr=None):
+    def __init__(
+        self, intervals: IntervalSet, force_repr: Optional[str] = None
+    ) -> None:
+        super().__init__()
         assert isinstance(intervals, IntervalSet)
         self.intervals = intervals
         self._force_repr = force_repr
@@ -72,6 +76,7 @@ class OneCharStringStrategy(SearchStrategy):
         )
         if codec is not None:
             intervals &= charmap.intervals_from_codec(codec)
+
         _arg_repr = ", ".join(
             f"{k}={v!r}"
             for k, v in [
@@ -82,7 +87,8 @@ class OneCharStringStrategy(SearchStrategy):
                 ("exclude_characters", exclude_characters),
                 ("include_characters", include_characters),
             ]
-            if v not in (None, "", set(charmap.categories()) - {"Cs"})
+            if v not in (None, "")
+            and not (k == "categories" and set(v) == set(charmap.categories()) - {"Cs"})
         )
         if not intervals:
             raise InvalidArgument(
@@ -116,10 +122,10 @@ class OneCharStringStrategy(SearchStrategy):
             f"{alphabet=} must be a sampled_from() or characters() strategy"
         )
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return self._force_repr or f"OneCharStringStrategy({self.intervals!r})"
 
-    def do_draw(self, data):
+    def do_draw(self, data: ConjectureData) -> str:
         return data.draw_string(self.intervals, min_size=1, max_size=1)
 
 
@@ -150,12 +156,12 @@ _nonempty_and_content_names = (
 )
 
 
-class TextStrategy(ListStrategy):
+class TextStrategy(ListStrategy[str]):
     def do_draw(self, data):
         # if our element strategy is OneCharStringStrategy, we can skip the
-        # ListStrategy draw and jump right to our nice IR string draw.
+        # ListStrategy draw and jump right to data.draw_string.
         # Doing so for user-provided element strategies is not correct in
-        # general, as they may define a different distribution than our IR.
+        # general, as they may define a different distribution than data.draw_string.
         elems = unwrap_strategies(self.element_strategy)
         if isinstance(elems, OneCharStringStrategy):
             return data.draw_string(
@@ -169,7 +175,7 @@ class TextStrategy(ListStrategy):
             )
         return "".join(super().do_draw(data))
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         args = []
         if repr(self.element_strategy) != "characters()":
             args.append(repr(self.element_strategy))
@@ -344,6 +350,7 @@ def _identifier_characters():
 
 class BytesStrategy(SearchStrategy):
     def __init__(self, min_size: int, max_size: Optional[int]):
+        super().__init__()
         self.min_size = min_size
         self.max_size = (
             max_size if max_size is not None else COLLECTION_DEFAULT_MAX_SIZE
